@@ -1,3 +1,4 @@
+import functools
 from functools import partial
 
 import torch
@@ -79,6 +80,9 @@ def batch_process(t, fn, chunks = 10, dim = 0):
     chunks = [fn(t_chunk) for t_chunk in t.chunk(chunks, dim = dim)]
     return torch.cat(chunks, dim = dim)
 
+def mult_reduce(arr):
+    return functools.reduce(lambda x, y: x * y, arr, 1)
+
 # gradient control
 
 def frac_gradient(t, frac):
@@ -390,7 +394,8 @@ class Sparse3DNA(nn.Module):
         dim_head = 64,
         dropout = 0.,
         causal = False,
-        query_num_frames_chunk = None
+        query_num_frames_chunk = None,
+        rel_pos_bias = False
     ):
         super().__init__()
         inner_dim = dim_head * heads
@@ -409,7 +414,11 @@ class Sparse3DNA(nn.Module):
         self.kernel_size = cast_tuple(kernel_size, size = 3)
         assert all(map(lambda n: n % 2 == 1, self.kernel_size)), 'kernel size must be odd'
 
-        self.kernel_numel = self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2]
+        self.kernel_numel = mult_reduce(self.kernel_size)
+
+        # relative positional bias per head, if needed
+
+        self.rel_pos_bias = AxialPositionalEmbedding(heads, shape = self.kernel_size) if rel_pos_bias else None
 
         # calculate padding
 
@@ -500,6 +509,14 @@ class Sparse3DNA(nn.Module):
         k, v = map(lambda t: rearrange(t, 'b (f h w) d -> b d f h w', f  = num_frames, h = fmap_size), (k, v))
         k, v = map(lambda t: F.pad(t, video_padding), (k, v))
 
+        # axial relative pos bias
+
+        rel_pos_bias = None
+
+        if exists(self.rel_pos_bias):
+            rel_pos_bias = rearrange(self.rel_pos_bias(), 'j h -> h 1 j')
+            rel_pos_bias = F.pad(rel_pos_bias, (1, 0), value = 0.)
+
         # put the attention processing code in a function
         # to allow for processing queries in chunks of frames
 
@@ -522,6 +539,11 @@ class Sparse3DNA(nn.Module):
 
             sim = einsum('b i d, b i j d -> b i j', q, k)
 
+            # add rel pos bias, if needed
+
+            if exists(rel_pos_bias):
+                sim = sim + rel_pos_bias
+
             # causal mask
 
             if exists(mask):
@@ -1061,6 +1083,7 @@ class Transformer(nn.Module):
         sparse_3dna_video_shape = None,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
+        sparse_3dna_rel_pos_bias = False,
         shift_video_tokens = False,
         rotary_pos_emb = False
     ):
@@ -1082,7 +1105,8 @@ class Transformer(nn.Module):
                     kernel_size = sparse_3dna_kernel_size,
                     dilation = dilation,
                     video_shape = sparse_3dna_video_shape,
-                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
+                    rel_pos_bias = sparse_3dna_rel_pos_bias,
                 )
             else:
                 self_attn = Attention(
@@ -1172,6 +1196,7 @@ class ReversibleTransformer(nn.Module):
         sparse_3dna_video_shape = None,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
+        sparse_3dna_rel_pos_bias = False,
         shift_video_tokens = False,
         rotary_pos_emb = False
     ):
@@ -1194,7 +1219,8 @@ class ReversibleTransformer(nn.Module):
                     kernel_size = sparse_3dna_kernel_size,
                     dilation = dilation,
                     video_shape = sparse_3dna_video_shape,
-                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
+                    rel_pos_bias = sparse_3dna_rel_pos_bias,
                 )
             else:
                 image_size = None
@@ -1638,18 +1664,33 @@ class AxialPositionalEmbedding(nn.Module):
         shape
     ):
         super().__init__()
-        self.dim = dim
-        frames, height, width = shape
-        self.pos_frames = nn.Parameter(torch.randn(frames, dim))
-        self.pos_height = nn.Parameter(torch.randn(height, dim))
-        self.pos_width = nn.Parameter(torch.randn(width, dim))
+        shape = tuple(filter(lambda t: t > 1, shape))
 
-    def forward(self):
-        pos_frames = rearrange(self.pos_frames, 'f d -> f 1 1 d')
-        pos_height = rearrange(self.pos_height, 'h d -> 1 h 1 d')
-        pos_width = rearrange(self.pos_width, 'w d -> 1 1 w d')
-        positions = pos_frames + pos_height + pos_width
-        return rearrange(positions, 'f h w d -> 1 (f h w) d')
+        self.dim = dim
+        self.shape = shape
+        self.num_axials = len(shape)
+
+        for axial_ind, axial_len in enumerate(shape):
+            axial_pos = nn.Parameter(torch.randn(axial_len, dim))
+            setattr(self, f'axial{axial_ind + 1}', axial_pos)
+
+    def forward(self, *, flatten = True):
+        positions = None
+
+        for axial_ind in range(self.num_axials):
+            axial_pos = getattr(self, f'axial{axial_ind + 1}')
+
+            if not exists(positions):
+                positions = axial_pos
+                continue
+
+            positions = rearrange(positions, '... d -> ... 1 d')
+            positions = positions + axial_pos
+
+        if flatten:
+            positions = rearrange(positions, '... d -> (...) d')
+
+        return positions
 
 # sampling helpers
 
@@ -1690,6 +1731,7 @@ class NUWA(nn.Module):
         sparse_3dna_kernel_size = 3,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilation = 1,
+        sparse_3dna_rel_pos_bias = False
     ):
         super().__init__()
         assert exists(vae) ^ exists(image_size), 'either VAE or image size must be specified'
@@ -1754,7 +1796,8 @@ class NUWA(nn.Module):
             sparse_3dna_attn = True,
             sparse_3dna_kernel_size = sparse_3dna_kernel_size,
             sparse_3dna_dilations = sparse_3dna_dilations,
-            sparse_3dna_query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+            sparse_3dna_query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
+            sparse_3dna_rel_pos_bias = sparse_3dna_rel_pos_bias
         )
 
         self.to_logits = nn.Linear(dim, num_image_tokens)
@@ -1818,7 +1861,7 @@ class NUWA(nn.Module):
                 video_indices_input = video_indices[:, -lookback_tokens:]
 
             frame_embeddings = self.image_embedding(video_indices_input)
-            frame_embeddings = pos_emb[:, :frame_embeddings.shape[1]] + frame_embeddings
+            frame_embeddings = pos_emb[:frame_embeddings.shape[1]] + frame_embeddings
             frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
 
             frame_embeddings = self.video_transformer(
@@ -1879,7 +1922,7 @@ class NUWA(nn.Module):
         frame_indices_input = frame_indices[:, :-1] if return_loss else frame_indices
 
         frame_embeddings = self.image_embedding(frame_indices_input)
-        frame_embeddings = self.video_pos_emb()[:, :-1] + frame_embeddings
+        frame_embeddings = self.video_pos_emb()[:-1] + frame_embeddings
 
         bos = repeat(self.video_bos, 'd -> b 1 d', b = batch)
         frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
@@ -1983,7 +2026,7 @@ class NUWAVideoAudio(nn.Module):
         self.audio_embedding = Embedding(num_audio_tokens, dim, frac_gradient = embed_gradient_frac)
 
         max_audio_seq_len = num_audio_tokens_per_video_frame * max_video_frames
-        self.audio_pos_emb = nn.Embedding(max_audio_seq_len, dim)
+        self.audio_pos_emb = AxialPositionalEmbedding(dim, shape = (num_audio_tokens // audio_tokens_per_timestep, audio_tokens_per_timestep))
 
         self.audio_loss_weight = audio_loss_weight
 
@@ -2093,13 +2136,13 @@ class NUWAVideoAudio(nn.Module):
             # prep video embeddings
 
             frame_embeddings = self.image_embedding(video_indices_input)
-            frame_embeddings = video_pos_emb[:, :frame_embeddings.shape[1]] + frame_embeddings
+            frame_embeddings = video_pos_emb[:frame_embeddings.shape[1]] + frame_embeddings
             frame_embeddings = torch.cat((video_bos, frame_embeddings), dim = 1)
 
             # prep audio embeddings
 
             audio_embeddings = self.audio_embedding(audio_indices_input)
-            audio_pos_emb = self.audio_pos_emb(torch.arange(audio_embeddings.shape[1], device = device))
+            audio_pos_emb = self.audio_pos_emb()[:audio_embeddings.shape[1]]
             audio_pos_emb = rearrange(audio_pos_emb, 'n d -> 1 n d')
             audio_embeddings = audio_embeddings + audio_pos_emb
             audio_embeddings = torch.cat((audio_bos, audio_embeddings), dim = 1)
@@ -2185,7 +2228,7 @@ class NUWAVideoAudio(nn.Module):
         frame_indices_input = frame_indices[:, :-1] if return_loss else frame_indices
 
         frame_embeddings = self.image_embedding(frame_indices_input)
-        frame_embeddings = self.video_pos_emb()[:, :-1] + frame_embeddings
+        frame_embeddings = self.video_pos_emb()[:-1] + frame_embeddings
 
         video_bos = repeat(self.video_bos, 'd -> b 1 d', b = batch)
         frame_embeddings = torch.cat((video_bos, frame_embeddings), dim = 1)
@@ -2195,7 +2238,7 @@ class NUWAVideoAudio(nn.Module):
         audio_indices_input = audio[:, :-1] if return_loss else audio
 
         audio_embeddings = self.audio_embedding(audio_indices_input)
-        audio_pos_emb = self.audio_pos_emb(torch.arange(audio_embeddings.shape[1], device = device))
+        audio_pos_emb = self.audio_pos_emb()[:audio_embeddings.shape[1]]
         audio_embeddings = audio_embeddings + rearrange(audio_pos_emb, 'n d -> 1 n d')
 
         audio_bos = repeat(self.audio_bos, 'd -> b 1 d', b = batch)
@@ -2360,7 +2403,7 @@ class NUWASketch(nn.Module):
         num_tokens = sketch_tokens.shape[1]
 
         sketch_pos_emb = self.sketch_pos_emb()
-        sketch_pos_emb = sketch_pos_emb[:, :num_tokens]
+        sketch_pos_emb = sketch_pos_emb[:num_tokens]
 
         sketch_tokens = sketch_tokens + sketch_pos_emb
 
@@ -2411,7 +2454,7 @@ class NUWASketch(nn.Module):
                 video_indices_input = video_indices[:, -lookback_tokens:]
 
             frame_embeddings = self.image_embedding(video_indices_input)
-            frame_embeddings = pos_emb[:, :frame_embeddings.shape[1]] + frame_embeddings
+            frame_embeddings = pos_emb[:frame_embeddings.shape[1]] + frame_embeddings
             frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
 
             frame_embeddings = self.video_transformer(
@@ -2482,7 +2525,7 @@ class NUWASketch(nn.Module):
         frame_indices_input = frame_indices[:, :-1] if return_loss else frame_indices
 
         frame_embeddings = self.image_embedding(frame_indices_input)
-        frame_embeddings = self.video_pos_emb()[:, :-1] + frame_embeddings
+        frame_embeddings = self.video_pos_emb()[:-1] + frame_embeddings
 
         bos = repeat(self.video_bos, 'd -> b 1 d', b = batch)
         frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
